%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%          Copyright (c) 2013 K. Neumann, CoR-Lab                     %%%
%%%          Bielefeld University, Germany, http://cor-lab.de           %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

classdef CELM < Learner
    %Constrained Least Squares ELM Extreme Learning Machine
    %   Single layer feed-forward network with constraints
    
    properties
        hidDim = 10;    %number of neurons in the hidden layer
        Winp = [];      %weights from input layer to hidden layer
        Wout = [];      %weights from hidden layer to output layer
        a = [];         %slope parameters of activation functions
        b = [];         %bias parameters of activation functions
        
        reg = 1e-3;     %regularization parameter for ridge regression
        flagBIP = 0;    %flag for using Batch Intrinsic Plasticity (BIP)
        mu = 0.2;       %desired mean activity parameter for BIP

        batchSize = 10000;
        bound = 1e10;   %output bound for Wout 
        vector = 0;
        
        normalization = 0;
        PrintIt = 2;
        
        CIEQCs = {};     %cell array with continuous ineqaulity constraints
        CEQCs = {};     %cell array with continuous eqaulity constraints
        DIEQCs = {};     %cell array with discrete ineqaulity constraints
        DEQCs = {};       %cell array with discrete eqaulity constraints
        ObjFct = {};
        maxIter = 1000; %maximal number of constraint-sampling iterations
        
        Ax = [];        %affine maxtrices of input transformation x' = Ax*x+bx
        bx = []; 
        Ay = [];        %affine maxtrices of output transformation y' = Ay*y+by
        by = []; 
        
        Aeq = [];       %equality constraints
        beq = [];       %here also
        Aieq = [];      %inequality constraints
        bieq = [];      %here also
        
        model = 'CELM';
    end
    
    methods
        function l = CELM(inpDim, outDim, spec)
            l = l@Learner(inpDim, outDim, spec);
        end
        
        function s = getDescriptorString(l)
            s = ['CELM' '-' 'hn' '-' num2str(l.hidDim) '-' 'reg' '-' num2str(l.reg)]; 
        end
        
        function init(l, X, Y)
            l.a = ones(1,l.hidDim);
            l.b = 2 * rand(1,l.hidDim) - ones(1,l.hidDim);
            l.Winp = 2 * rand(l.inpDim,l.hidDim) - ones(l.inpDim,l.hidDim);
            l.Wout = 2 * rand(l.hidDim,l.outDim) - ones(l.hidDim,l.outDim);    
            
            % setting the test values for the continuous inequality constraints
            num_objfct = length(l.ObjFct);
            for i = 1:num_objfct
                l.ObjFct{i}.inpDim = l.inpDim;
            end
            
            num_dieqs = length(l.DIEQCs);
            for i = 1:num_dieqs
                l.DIEQCs{i}.inpDim = l.inpDim;
            end
            
            num_deqs = length(l.DEQCs);
            for i = 1:num_deqs
                l.DEQCs{i}.inpDim = l.inpDim;
            end
            
            % setting the test values for the continuous inequality constraints
            num_cieqcs = length(l.CIEQCs);
            for i = 1:num_cieqcs
                l.CIEQCs{i}.inpDim = l.inpDim;
                if l.CIEQCs{i}.maxTestValue == 0 
                    l.CIEQCs{i}.maxTestValue = l.CIEQCs{i}.maxValue;
                end
                if l.CIEQCs{i}.minTestValue == 0 
                    l.CIEQCs{i}.minTestValue = l.CIEQCs{i}.minValue;
                end
            end
            
            % setting the test values for the continuous equality constraints
            num_ceqcs = length(l.CEQCs);
            for i = 1:num_ceqcs
                l.CEQCs{i}.inpDim = l.inpDim;
                if l.CEQCs{i}.maxTestValue == 0 
                    l.CEQCs{i}.maxTestValue = l.CEQCs{i}.maxValue;
                end
                if l.CEQCs{i}.minTestValue == 0 
                    l.CEQCs{i}.minTestValue = l.CEQCs{i}.minValue;
                end
            end
            
            if isempty(l.ObjFct) == 0
                l.normalization = 0;
            end
            
            if l.normalization
                l.norm(X,Y);                        
            end 
            
            if l.flagBIP
                if l.normalization == 1
                    [X,~] = l.normalizeIO(X,Y);
                end
                l.bip(X);
            end
        end
        
        function ClearConstraints(l)
           l.Aieq = cell(1,l.outDim); 
           l.bieq = cell(1,l.outDim); 
           l.Aeq = cell(1,l.outDim); 
           l.beq = cell(1,l.outDim); 
        end
        
        function IEQConstraints(l,Cons)
            a = l.a'; b = l.b'; Winp = l.Winp'; 
            num_cc = size(Cons,2);
            for c = 1:num_cc
                uc = Cons{c}.u';
                USize = size(uc,2);
                if USize ~= 0
                    num_parts = size(Cons{c}.partial,2);
                    PartialDerivs = Cons{c}.partial;
                    Coeffs = Cons{c}.factors;
                    for pd = 1:num_parts
                        AieqMinTemp = []; bieqMinTemp = 0; AieqMaxTemp = []; bieqMaxTemp = 0;
                        partials = PartialDerivs{pd};
                        factors = Coeffs{pd};
                        if strcmp(class(factors),'function_handle') == 1
                            factors = factors(Cons{c}.u);
                        else
                            factors = repmat(factors,USize,1);
                        end
                        for s = 1:l.outDim
                            if Cons{c}.minValue ~= -Inf
                                D = size(partials{s},2);
                                matTemp = ones(l.hidDim,USize);
                                for p = partials{s}
                                    WinpAx = (Winp*l.Ax);
                                    matTemp = matTemp.*repmat(a.*WinpAx(:,p),1,USize);
                                end   
                                if D == 0
                                    AieqMinTemp = [AieqMinTemp,-inv(l.Ay(s,s))*repmat(factors(:,s),1,l.hidDim).*fermi(diag(a)*Winp*(l.Ax*uc+repmat(l.bx,1,USize))+repmat(b,1,USize))'];
                                    bieqMinTemp = bieqMinTemp + repmat(-Cons{c}.minValue-l.Ay(s,s)\l.by(s),USize,1); %Glg überdenken
                                else
                                    AieqMinTemp = [AieqMinTemp,-inv(l.Ay(s,s))*repmat(factors(:,s),1,l.hidDim).*(calcFermiDerivative(D,(diag(a)*Winp)*(l.Ax*uc+repmat(l.bx,1,USize))+repmat(b,1,USize)) .* matTemp)'];
                                    bieqMinTemp = bieqMinTemp + repmat(-Cons{c}.minValue,USize,1); %Glg überdenken
                                end
                            else
                                bieqMinTemp = [];
                            end
                            if Cons{c}.maxValue ~= Inf
                                D = size(partials{s},2);
                                matTemp = ones(l.hidDim,USize);
                                for p = partials{s}
                                    WinpAx = (Winp*l.Ax);
                                    matTemp = matTemp.*repmat(a.*WinpAx(:,p),1,USize);
                                end
                                if D == 0
                                    AieqMaxTemp = [AieqMaxTemp,inv(l.Ay(s,s))*repmat(factors(:,s),1,l.hidDim).*fermi(diag(a)*Winp*(l.Ax*uc+repmat(l.bx,1,USize))+repmat(b,1,USize))'];
                                    bieqMaxTemp = bieqMaxTemp + repmat(Cons{c}.maxValue+l.Ay(s,s)\l.by(s),USize,1); %Glg überdenken
                                else
                                    AieqMaxTemp = [AieqMaxTemp,inv(l.Ay(s,s))*repmat(factors(:,s),1,l.hidDim).*(calcFermiDerivative(D,(diag(a)*Winp)*(l.Ax*uc+repmat(l.bx,1,USize))+repmat(b,1,USize)) .* matTemp)'];
                                    bieqMaxTemp = bieqMaxTemp + repmat(Cons{c}.maxValue,USize,1); %Glg überdenken
                                end
                            else
                                bieqMaxTemp = [];
                            end
                        end
                        if pd == 1
                            AieqTempSum = [AieqMinTemp;AieqMaxTemp]; 
                            bieqTempSum = [bieqMinTemp;bieqMaxTemp];    
                        else
                            AieqTempSum = AieqTempSum + [AieqMinTemp;AieqMaxTemp]; 
                            bieqTempSum = bieqTempSum + [bieqMinTemp;bieqMaxTemp];
                        end
                    end            
                    l.Aieq{1} = [l.Aieq{1};AieqTempSum]; l.bieq{1} = [l.bieq{1};bieqTempSum];
                end
            end
        end
        
        function EQConstraints(l,Cons)
            a = l.a'; b = l.b';
            Wout = l.Wout'; Winp = l.Winp';
            
            if isempty(l.Ax)
                l.Ax = eye(l.inpDim);
            end
            if isempty(l.bx)
                l.bx = zeros(l.inpDim,1);
            end
            if isempty(l.Ay)
                l.Ay = eye(l.outDim);
            end
            if isempty(l.by)
                l.by = zeros(l.outDim,1);
            end
            
            num_cc = size(Cons,2);
            AeqTempSum = []; beqTempSum = [];
            for c = 1:num_cc
                uc = Cons{c}.u';
                USize = size(uc,2);
                if USize == 0
                    return;
                end
                num_parts = size(Cons{c}.partial,2);
                PartialDerivs = Cons{c}.partial;
                Coeffs = Cons{c}.factors;
                for pd = 1:num_parts
                    AeqTemp = []; beqTemp = 0;
                    partials = PartialDerivs{pd};
                    factors = Coeffs{pd};
                    if strcmp(class(factors),'function_handle') == 1
                        factors = factors(uc);
                    else
                        factors = repmat(factors,USize,1);
                    end
                    for s = 1:l.outDim
                        D = size(partials{s},2);
                        matTemp = ones(l.hidDim,USize);
                        for p = partials{s}
                            WinpAx = (Winp*l.Ax);
                            matTemp = matTemp.*repmat(a.*WinpAx(:,p),1,USize);
                        end   
                        if D == 0
                            AeqTemp = [AeqTemp,-inv(l.Ay(s,s))*repmat(factors(:,s),1,l.hidDim).*fermi(diag(a)*Winp*(l.Ax*uc+repmat(l.bx,1,USize))+repmat(b,1,USize))'];
                            beqTemp = beqTemp + repmat(-Cons{c}.Value-l.Ay(s,s)\l.by(s),USize,1); %Glg überdenken
                        else
                            AeqTemp = [AeqTemp,-inv(l.Ay(s,s))*repmat(factors(:,s),1,l.hidDim).*(calcFermiDerivative(D,(diag(a)*Winp)*(l.Ax*uc+repmat(l.bx,1,USize))+repmat(b,1,USize)) .* matTemp)'];
                            beqTemp = beqTemp + repmat(-Cons{c}.Value,USize,1);  %Glg überdenken
                        end
                    end
                    if pd == 1
                        AeqTempSum = AeqTemp; 
                        beqTempSum = beqTemp;    
                    else
                        AeqTempSum = AeqTempSum + AeqTemp; 
                        beqTempSum = beqTempSum + beqTemp;
                    end
                end
                l.Aeq{1} = [l.Aeq{1};AeqTempSum]; l.beq{1} = [l.beq{1};beqTempSum];
            end
        end
        
        function [Heff,Yeff] = ObjectiveFunction(l,X,Y)
            a = l.a'; b = l.b';
            Wout = l.Wout'; Winp = l.Winp';
            
            H = []; XT = X'; Heff = []; Yeff = [];
            
            if isempty(l.ObjFct)==1
                 Heff = l.calcHiddenStates(X);
                 Yeff = Y;
                 return;
            end

            num_cc = size(l.ObjFct,2);
            for c = 1:num_cc
                XSize = size(XT,2);
                if XSize == 0
                    return;
                end
                num_parts = size(l.ObjFct{c}.partial,2);
                PartialDerivs = l.ObjFct{c}.partial;
                Coeffs = l.ObjFct{c}.factors;
                for pd = 1:num_parts
                    HTemp = [];
                    partials = PartialDerivs{pd};
                    factors = Coeffs{pd};
                    if strcmp(class(factors),'function_handle') == 1
                        factors = factors(X);
                    else
                        factors = repmat(factors,XSize,1);
                    end
                    for s = 1:l.outDim
                        D = size(partials{s},2);
                        matTemp = ones(l.hidDim,XSize);
                        for p = partials{s}
                            WinpAx = (Winp*l.Ax);
                            matTemp = matTemp.*repmat(a.*WinpAx(:,p),1,XSize);
                        end   
                        if D == 0
                            HTemp = [HTemp,repmat(factors(:,s),1,l.hidDim).*fermi(diag(a)*Winp*XT+repmat(b,1,XSize))'];
                        else
                            HTemp = [HTemp,repmat(factors(:,s),1,l.hidDim).*(calcFermiDerivative(D,(diag(a)*Winp)*XT+repmat(b,1,XSize)) .* matTemp)'];
                        end
                    end
                    if pd == 1
                        H = HTemp;
                    else
                        H = H + HTemp;
                    end
                end
                Heff = [Heff;H];
                Yeff = [Yeff;Y(:,c)];
            end
        end
        
        function ConstraintLeastSquares(l, X, Y)
            options = optimset('lsqlin');
            options = optimset(options,'Algorithm','interior-point');
%             options = optimset(options,'Algorithm','trust-region-reflective');
            options = optimset(options,'maxIter',1e9);
            options = optimset(options,'LargeScale','on','Display','off');
            %options = optimset(options,'Display','off');
    
            if l.normalization == 1
                X = l.range2norm(X, l.inpRange, l.inpOffset);
                Y = l.range2norm(Y, l.outRange, l.outOffset);
            end
            WoutTemp = reshape(l.Wout,l.outDim*l.hidDim,1);
            
            [H,Y] = l.ObjectiveFunction(X,Y);

            [xH,yH] = size(H);
            lb = -l.bound*ones(1,l.hidDim*l.outDim); ub = l.bound*ones(1,l.hidDim*l.outDim);
            
            C = []; d = [];
            for i = 1:l.outDim
                C = [C;[zeros(xH,(i-1)*yH),H,zeros(xH,(l.outDim-i)*yH)]];
                d = [d;Y(:,i)];
            end
            
            %Tikhonov Regularization
            C = [C;sqrt(l.reg)*eye(l.hidDim*l.outDim)]; d = [d;zeros(l.hidDim*l.outDim,1)];
            
            %Constrained Regression   
            x = lsqlin(sparse(C),d,l.Aieq{1},l.bieq{1},l.Aeq{1},l.beq{1},lb,ub,[],options);
           
            %Assignment
            l.Wout = reshape(x',l.hidDim,l.outDim);
        end
        
        function train(l, X, Y)
            XT = X'; YT = Y';  
                        
            % number of continuous constraints
            num_cieqcs = size(l.CIEQCs,2);
            num_ceqcs = size(l.CEQCs,2);
            
            % ridge regression
            %l.trainRR(X,Y);
            
            l.ClearConstraints();
            l.IEQConstraints(l.DIEQCs);
            l.EQConstraints(l.DEQCs);
            l.ConstraintLeastSquares(X,Y);
            
            if num_cieqcs + num_ceqcs ~= 0

                for i = 1:l.maxIter
                    
                    %constraint fulfillment check
                    uTestSamples_cieqcs = cell(1,num_cieqcs); uTestSamples_ceqcs = cell(1,num_ceqcs);
                    for c = 1:num_cieqcs
                        uTestSamples_cieqcs{c} = l.CIEQCs{c}.DrawTestSamples();
                    end
                    for c = 1:num_ceqcs
                        uTestSamples_ceqcs{c} = l.CEQCs{c}.DrawTestSamples();
                    end

                    violationcount_cieqcs = zeros(1,num_cieqcs); violationcount_ceqcs = zeros(1,num_ceqcs);
                    for c = 1:num_cieqcs
                        num_parts = size(l.CIEQCs{c}.partial,2);
                        PartialDerivs = l.CIEQCs{c}.partial;
                        Coeffs = l.CIEQCs{c}.factors;
                        val = zeros(l.CIEQCs{c}.TestSamplesPerIteration,1);
                        for p = 1:num_parts
                            partials = PartialDerivs{p};
                            factors = Coeffs{p};
                            val = val + l.fctOutput(uTestSamples_cieqcs{c},partials,factors);
                        end
                        violationcount_cieqcs(c) = sum(val>l.CIEQCs{c}.maxTestValue|val<l.CIEQCs{c}.minTestValue);
                    end
                    for c = 1:num_ceqcs
                        num_parts = size(l.CEQCs{c}.partial,2);
                        PartialDerivs = l.CEQCs{c}.partial;
                        Coeffs = l.CEQCs{c}.factors;
                        val = zeros(l.CIEQCs{c}.TestSamplesPerIteration,1);
                        for p = 1:num_parts
                            partials = PartialDerivs{p};
                            factors = Coeffs{p};
                            val = val + l.fctOutput(uTestSamples_ceqcs{c},partials,factors);
                        end
                        violationcount_ceqcs(c) = sum(val>l.CEQCs{c}.maxTestValue|val<l.CEQCs{c}.minTestValue);
                    end
                    
                    %printing
                    if l.PrintIt == 1
                        fprintf('*');
                    end
                    if l.PrintIt == 2
                        fprintf('Iter.: %d, ',i);
                        if num_cieqcs > 0
                            for c = 1:num_cieqcs-1
                                fprintf('Rel. (%d): %2.3f, ',size(l.CIEQCs{c}.u,1),1-(violationcount_cieqcs(c)/l.CIEQCs{c}.TestSamplesPerIteration));
                            end
                            fprintf('Rel. (%d): %2.3f, ',size(l.CIEQCs{num_cieqcs}.u,1),1-(violationcount_cieqcs(num_cieqcs)/l.CIEQCs{num_cieqcs}.TestSamplesPerIteration));
                        end
                        
                        if num_ceqcs > 0
                            for c = 1:num_ceqcs-1
                                fprintf('Rel. (%d): %2.3f, ',size(l.CEQCs{c}.u,1),1-(violationcount_ceqcs(c)/l.CEQCs{c}.TestSamplesPerIteration));
                            end
                            fprintf('Rel. (%d): %2.3f\n',size(l.CEQCs{num_ceqcs}.u,1),1-(violationcount_ceqcs(num_ceqcs)/l.CEQCs{num_ceqcs}.TestSamplesPerIteration));
                        else
                            fprintf('\n');
                        end
                    end
                    
                    TestSamplesPerIteration_cieqcs = zeros(1,num_cieqcs); StabilityThresholds_cieqcs = zeros(1,num_cieqcs);
                    for c = 1:num_cieqcs
                        TestSamplesPerIteration_cieqcs(c) = l.CIEQCs{c}.TestSamplesPerIteration;
                        StabilityThresholds_cieqcs(c) = l.CIEQCs{c}.StabilityThreshold;
                    end
                    TestSamplesPerIteration_ceqcs = zeros(1,num_ceqcs); StabilityThresholds_ceqcs = zeros(1,num_ceqcs);
                    for c = 1:num_ceqcs
                        TestSamplesPerIteration_ceqcs(c) = l.CEQCs{c}.TestSamplesPerIteration;
                        StabilityThresholds_ceqcs(c) = l.CEQCs{c}.StabilityThreshold;
                    end
                    
                    %break test
                    flag_StopSampling_cieqcs = 1-(violationcount_cieqcs./TestSamplesPerIteration_cieqcs) > StabilityThresholds_cieqcs;
                    flag_StopSampling_ceqcs = 1-(violationcount_ceqcs./TestSamplesPerIteration_ceqcs) > StabilityThresholds_ceqcs;
                    if sum(flag_StopSampling_cieqcs) == num_cieqcs && sum(flag_StopSampling_ceqcs) == num_ceqcs
                        break;
                    end

                    for c = 1:num_cieqcs
                        if flag_StopSampling_cieqcs(c) == 0
                            num_parts = size(l.CIEQCs{c}.partial,2);
                            PartialDerivs = l.CIEQCs{c}.partial;
                            Coeffs = l.CIEQCs{c}.factors;
                            val = zeros(l.CIEQCs{c}.TestSamplesPerIteration,1);
                            for p = 1:num_parts
                                partials = PartialDerivs{p};
                                factors = Coeffs{p};
                                val = val + l.fctOutput(uTestSamples_cieqcs{c},partials,factors);
                            end
                            
                            val_pos = val - l.CIEQCs{c}.maxTestValue;
                            val_neg = -(val - l.CIEQCs{c}.minTestValue); %ABS FCT anwenden?
                            val_comb = max(val_pos,val_neg);
                            [sortval,sortind] = sort(val_comb, 'descend');
                            
                            j = 1; counter = 0;
                            uTestSamplesTemp = uTestSamples_cieqcs{c}(sortind,:);
                           
                            while j < size(uTestSamplesTemp,1) && counter < l.CIEQCs{c}.SamplesPerIteration
                                if sortval(j) > 0
                                    l.CIEQCs{c}.u = [l.CIEQCs{c}.u;uTestSamplesTemp(j,:)];
                                    counter = counter + 1;
                                end
                                j = j + 1;
                            end
                        end
                    end
                    for c = 1:num_ceqcs
                        if flag_StopSampling_ceqcs(c) == 0
                            num_parts = size(l.CEQCs{c}.partial,2);
                            PartialDerivs = l.CEQCs{c}.partial;
                            Coeffs = l.CEQCs{c}.factors;
                            val = zeros(l.CEQCs{c}.TestSamplesPerIteration,1);
                            for p = 1:num_parts
                                partials = PartialDerivs{p};
                                factors = Coeffs{p};
                                val = val + l.fctOutput(uTestSamples_ceqcs{c},partials,factors);
                            end

                            val_pos = val - l.CEQCs{c}.maxTestValue;
                            val_neg = -(val - l.CEQCs{c}.minTestValue); %ABS FCT anwenden?
                            val_comb = max(val_pos,val_neg);
                            [sortval,sortind] = sort(val_comb, 'descend');

                            j = 1; counter = 0;
                            uTestSamplesTemp = uTestSamples_ceqcs{c}(sortind,:);
                            while j < size(uTestSamplesTemp,1) && counter < l.CEQCs{c}.SamplesPerIteration
                                if sortval(j) > 0
                                    l.CEQCs{c}.u = [l.CEQCs{c}.u;uTestSamplesTemp(j,:)];
                                    counter = counter + 1;
                                end
                                j = j + 1;
                            end
                        end
                    end

                    % constraint learning                    
                    l.ClearConstraints();
                    l.IEQConstraints(l.CIEQCs);
                    l.EQConstraints(l.CEQCs);
                    l.IEQConstraints(l.DIEQCs);
                    l.EQConstraints(l.DEQCs);
                    l.ConstraintLeastSquares(X,Y);
    %                 Wout = l.Wout';
    %                 US = size(l.u{1},2); a = l.a'; b = l.b';
    %                 vvals = inv(l.Ay)*(Wout*calcFermiDerivative(0,(repmat(a,1,l.inpDim).*Winp)*(l.Ax*l.u{1}+repmat(l.bx,1,US))+repmat(b,1,US)) - l.by);
    %                 vvals2 = Wout*inv(l.Ay)*fermi(diag(a)*Winp*(l.Ax*l.u{1}+repmat(l.bx,1,US))+repmat(b,1,US)) - inv(l.Ay)*l.by;
    %                 vvals3 = l.apply(l.u{1}');
    %                 
    %                 US = size(l.u{2},2);
    %                 matTemp = ones(l.hidDim,US);
    %                 for p = l.partial{c}
    %                     WinpAx = (Winp*l.Ax);
    %                     matTemp = matTemp.*repmat(a.*WinpAx(:,p),1,US);
    %                 end
    %                 vvalsd = inv(l.Ay)*Wout*(calcFermiDerivative(D,(repmat(a,1,l.inpDim).*Winp)*(l.Ax*l.u{2}+repmat(l.bx,1,US))+repmat(b,1,US)) .* matTemp);


                end
            else
                l.ClearConstraints();
                l.IEQConstraints(l.DIEQCs);
                l.EQConstraints(l.DEQCs);
                l.ConstraintLeastSquares(X,Y);
            end
            
            % printing
            fprintf('Learning finished\n');
            
        end
        
        function val = fctOutput(l,X,partials,factors)
            XT = X'; 
            Winp = l.Winp'; 
            Wout = l.Wout';
            a = l.a';
            b = l.b';
            
            [NumSamples,~] = size(X);
            outDim = size(partials,2);
            for s = 1:outDim
                D = size(partials{s},2); 

                matTemp = ones(l.hidDim,NumSamples);
                for p = partials{s}
                    WinpAx = (Winp*l.Ax);
                    z = WinpAx(:,p);
                    matTemp = matTemp.*repmat(a.*WinpAx(:,p),1,NumSamples);
                end
                if D == 0
                    valTemp = inv(l.Ay)*(Wout*fermi(diag(a)*Winp*(l.Ax*XT+repmat(l.bx,1,NumSamples))+repmat(b,1,NumSamples)) - repmat(l.by,1,NumSamples));
                else
                    valTemp = inv(l.Ay)*Wout*(calcFermiDerivative(D,(repmat(a,1,l.inpDim).*Winp)*(l.Ax*XT+repmat(l.bx,1,NumSamples))+repmat(b,1,NumSamples)) .* matTemp);
                end
%                 val(s,:) = factors(s)(XT)*valTemp(s,:); %TODO
                if strcmp(class(factors),'function_handle') == 1
                    facts = factors(X);
                else
                    facts = repmat(factors,NumSamples,1);
                end
                val(s,:) = facts(:,s)'.*valTemp(s,:);
            end
            val = sum(val,1)';
        end
        
        function trainRR(l, X, Y, W)   
            if l.normalization == 1
                X = l.range2norm(X,l.inpRange,l.inpOffset);
                Y = l.range2norm(Y,l.outRange,l.outOffset);
            end
            if size(X,1) < l.batchSize
                [hs,Ys] = l.ObjectiveFunction(X,Y);
                if nargin < 4
                    l.Wout = ((hs'*hs + l.reg * eye(l.hidDim))\(hs' * Ys));
                else
                    WS = diag(W);
                    l.Wout = ((hs'*WS*hs + l.reg * eye(l.hidDim))\(hs' *WS* Ys));
                end
            else
                HTH = zeros(l.hidDim, l.hidDim);
                HTY = zeros(l.hidDim, l.outDim);                
                numBatches = floor(size(X,1)/l.batchSize);
                rest = mod(size(X,1),l.batchSize);                
                for bas=1:numBatches
                    [H,YS] = l.ObjectiveFunction(X((bas-1)*l.batchSize+1:bas*l.batchSize,:),Y((bas-1)*l.batchSize+1:bas*l.batchSize,:));
                    if nargin < 4
                        HTH = HTH + H' * H;
                        HTY = HTY + H' * YS;
                    else
                        WS = diag(W((bas-1)*l.batchSize+1:bas*l.batchSize));
                        HTH = HTH + H' * WS * H;
                        HTY = HTY + H' * WS * YS;
                    end
                end
                if rest > 0
                    [H,YS] = l.ObjectiveFunction(X(numBatches*l.batchSize+1:end,:),Y(numBatches*l.batchSize+1:end,:));
                    if nargin < 4
                        HTH = HTH + H' * H;
                        HTY = HTY + H' * YS;
                    else
                        WS = diag(W(numBatches*l.batchSize+1:end,:));
                        HTH = HTH + H' * WS * H;
                        HTY = HTY + H' * WS * YS;
                    end
                end                
                HTH = HTH + diag(repmat(l.reg,1,l.hidDim));
                l.Wout = (HTH\HTY);
            end
        end
        
        function [Y H] = apply(l, X)
            if iscell(X)
                Y = cell(length(X),1);
                for i=1:length(X)
                    Y{i} = l.apply(X{i});
                end
            else
                if l.normalization == 1
                    X = l.range2norm(X, l.inpRange, l.inpOffset);
                end
                if size(X,1) < l.batchSize
                    H = l.calcHiddenStates(X);
                    Y = H * l.Wout;
                else
                    Y = zeros(size(X,1), l.outDim);
                    numBatches = floor(size(X,1)/l.batchSize);
                    rest = mod(size(X,1),l.batchSize);
                    for b=1:numBatches
                        H = l.calcHiddenStates(X((b-1)*l.batchSize+1:b*l.batchSize,:));
                        Y((b-1)*l.batchSize+1:b*l.batchSize,:) = H * l.Wout;
                    end
                    if rest > 0
                        H = l.calcHiddenStates(X(numBatches*l.batchSize+1:end,:));
                        Y(numBatches*l.batchSize+1:end,:) = H * l.Wout;
                    end
                end
                if l.normalization == 1
                    Y = l.norm2range(Y, l.outRange, l.outOffset);
                end
                l.out = Y(end,:);
            end
        end
        
        function [H G] = calcHiddenStates(l, X)
            XT = X';
            Winp = l.Winp'; 
            Wout = l.Wout';
            a = l.a';
            b = l.b';
            
            numSamples = size(XT,2);
            atemp = repmat(a, 1, numSamples);
            btemp = repmat(b, 1, numSamples);
            G = Winp * XT;
            H = 1./(1+exp(-atemp .* G - btemp));
            H = H';
        end
        
        function bip(l, X)
            numSamples = size(X,1);
            
            PD = ProbDistUnivParam('exponential', l.mu);
            G = X * l.Winp;

            for hn = 1:l.hidDim
                targets = random(PD, 1, numSamples);

                hightars = length([targets(targets>=1),targets(targets<=0)]);
                while hightars > 0
                    further_targets = random(PD, 1, hightars);
                    targets = [targets(targets<1&targets>0),further_targets];
                    hightars = length([targets(targets>=1),targets(targets<=0)]);
                end
                targets = sort(targets);
                
                s = sort(G(:,hn));
                Phi = [s,ones(numSamples,1)];
                targetsinv = -log(1./targets - 1); %apply inverse activation function
                
                w = pinv(Phi)*targetsinv';                
                
                l.a(hn) = w(1);
                l.b(hn) = w(2);
            end
        end
        function norm(l, X, Y)
            XT = X; YT = Y;
            if iscell(X)
                [X, Y] = l.normalizeIO(cell2mat(X), cell2mat(Y));
            else
                [X, Y] = l.normalizeIO(X, Y);
                l.Ax = diag(2./(l.inpRange(2,:)-l.inpRange(1,:)));
                l.bx = -(2./(l.inpRange(2,:)-l.inpRange(1,:))).*(l.inpOffset + l.inpRange(1,:)) - 1;
                l.Ay = diag(2./(l.outRange(2,:)-l.outRange(1,:)));
                l.by = -(2./(l.outRange(2,:)-l.outRange(1,:))).*(l.outOffset + l.outRange(1,:)) - 1;
            end
            l.bx = l.bx';
            l.by = l.by';
            
            if l.normalization == 0
               l.Ax = eye(l.inpDim);
               l.bx = zeros(l.inpDim,1);
               l.Ay = eye(l.outDim);
               l.by = zeros(l.outDim,1);
            end
            
%             range2norm(0.5,l.inpRange,l.inpOffset)
%             l.Ax*0.5+l.bx
%             
%             (l.Ax*XT'+repmat(l.bx,1,length(l.Ax*XT')))' %=X
%             (l.Ay*YT'+repmat(l.by,1,length(l.Ay*YT')))' %=Y 
        end
        function [X, Y] = normalizeIO(l, X, Y)
            [X, l.inpOffset, l.inpRange] = l.normalize(X);
            if exist('Y', 'var')
                [Y, l.outOffset, l.outRange] = l.normalize(Y);
            end
        end
        function [data offset range] = normalize(l,data)
            %NORMALIZE Summary of this function goes here
            %   Detailed explanation goes here

            offset = mean(data, 1);
            data = data - repmat(offset, size(data,1), 1);

            range = [min(data, [], 1); max(data, [], 1)];
            data = l.range2norm(data, range);
        end
        function [y] = norm2range(l, x, ra, offset, minmax)
            %NORM2RANGE y = norm2range(x, ra, minmax)
            %   ra is a dim x 2 matrix holding min and max values per dimension
            %   minmax is repmat([-1 1], dim, 1) by default

            if nargin < 4
                offset = zeros(1, size(x,2));
            end

            if nargin < 5
                minmax = repmat([-1 1]', 1, size(x,2));
            end

            %cut exceeding dimensions
            ra = ra(:,1:min(size(ra,2), size(x,2)));
            x = x(:,1:min(size(ra,2), size(x,2)));

            tmp = minmax(2,:) - minmax(1,:);
            tmp(tmp==0) = 1;
            scale = 1./(tmp);
            range = (ra(2,:) - ra(1,:));

            y = zeros(size(x));
            for i=1:size(x,1)
                y(i,:) = (scale .* (x(i,:) - minmax(1,:))) .* range + ra(1,:);
                y(i,:) = y(i,:) + offset;
            end
        end
        function [y] = range2norm(l, x, ra, offset, minmax)
            %RANGE2NORM y = range2norm(x, ra, minmax)
            %   ra is a dim x 2 matrix holding min and max values per dimension
            %   minmax is repmat([-1 1], dim, 1) by default

            if nargin < 4
                offset = zeros(1, size(x,2));
            end

            if nargin < 5
                minmax = repmat([-1 1]', 1, size(x,2));
            end

            %cut exceeding dimensions
            ra = ra(:,1:min(size(ra,2), size(x,2)));
            x = x(:,1:min(size(ra,2), size(x,2)));

            tmp = ra(2,:) - ra(1,:);
            tmp(tmp==0) = 1;
            scale = 1 ./ (tmp);

            y = zeros(size(x));
            for i=1:size(x,1)
                y(i,:) = x(i,:) - offset;
                %y(i,:) = 2 * (((y(i,:) - ra(1,:)) .* scale) - 0.5);
                y(i,:) = ((minmax(2,:) - minmax(1,:)) .* ((y(i,:) - ra(1,:)) .* scale)) + minmax(1,:);
            end
        end
    end
    
end

%         function CIEQConstraints_IndOut(l)
%             a = l.a'; b = l.b';
%             Wout = l.Wout'; Winp = l.Winp';
%             
%             num_cc = size(l.CIEQCs,2);
%             for c = 1:num_cc
%                 partials = l.CIEQCs{c}.partial;
%                 USize = size(l.CIEQCs{c}.u,2);
%                 if USize ~= 0
%                     for s = 1:l.outDim
%                         if strcmp(class(l.CIEQCs{c}.factors),'function_handle') == 1
%                             factors = l.CIEQCs{c}.factors(l.CIEQCs{c}.u);
%                         else
%                             factors = l.CIEQCs{c}.factors;
%                         end
%                         if sum(factors(:,s)) ~= 0 %TODO
%                             if l.CIEQCs{c}.minValue ~= -Inf
%                                 D = size(partials{s},2);
%                                 matTemp = ones(l.hidDim,USize);
%                                 for p = partials{s}
%                                     WinpAx = (Winp*l.Ax);
%                                     matTemp = matTemp.*repmat(a.*WinpAx(:,p),1,USize);
%                                 end
%                                 if D == 0
%                                     l.Aieq{s} = [l.Aieq{s};-inv(l.Ay(s,s))*factors(:,s).*fermi(diag(a)*Winp*(l.Ax*l.CIEQCs{c}.u+repmat(l.bx,1,USize))+repmat(b,1,USize))'];
%                                     l.bieq{s} = [l.bieq{s};repmat(-l.CIEQCs{c}.minValue-l.Ay(s,s)\l.by(s),USize,1)];
%                                 else
%                                     l.Aieq{s} = [l.Aieq{s};-inv(l.Ay(s,s))*factors(:,s).*(calcFermiDerivative(D,(diag(a)*Winp)*(l.Ax*l.CIEQCs{c}.u+repmat(l.bx,1,USize))+repmat(b,1,USize)) .* matTemp)'];
%                                     l.bieq{s} = [l.bieq{s};repmat(-l.CIEQCs{c}.minValue,USize,1)];
%                                 end
%                             end
%                             if l.CIEQCs{c}.maxValue ~= Inf
%                                 D = size(partials{s},2);
%                                 matTemp = ones(l.hidDim,USize);
%                                 for p = partials{s}
%                                     WinpAx = (Winp*l.Ax);
%                                     matTemp = matTemp.*repmat(a.*WinpAx(:,p),1,USize);
%                                 end
%                                 if D == 0
%                                     l.Aieq{s} = [l.Aieq{s};inv(l.Ay(s,s))*factors(:,s).*fermi(diag(a)*Winp*(l.Ax*l.CIEQCs{c}.u+repmat(l.bx,1,USize))+repmat(b,1,USize))'];
%                                     l.bieq{s} = [l.bieq{s};repmat(l.CIEQCs{c}.maxValue+l.Ay(s,s)\l.by(s),USize,1)];
%                                 else
%                                     l.Aieq{s} = [l.Aieq{s};inv(l.Ay(s,s))*factors(:,s).*(calcFermiDerivative(D,(diag(a)*Winp)*(l.Ax*l.CIEQCs{c}.u+repmat(l.bx,1,USize))+repmat(b,1,USize)) .* matTemp)'];
%                                     l.bieq{s} = [l.bieq{s};repmat(l.CIEQCs{c}.maxValue,USize,1)];
%                                 end
% 
%                             end
%                         end
%                     end
%                 end
%             end
%         end
% 
%         function constraintleastsquares_IndOut(l, X, Y)
%             options = optimset('lsqlin');
%             options = optimset(options,'maxIter',1e8);
%             options = optimset(options,'LargeScale','off');
%             options = optimset(options,'Display','off');
%     
%             if l.normalization == 1
%                 X = l.range2norm(X, l.inpRange, l.inpOffset);
%                 Y = l.range2norm(Y, l.outRange, l.outOffset);
%             end
%             targets = Y';                        
%             H = l.calcHiddenStates(X);
%             lb = -l.bound*ones(1,l.hidDim); ub = l.bound*ones(1,l.hidDim);
%             
%             WoutTemp = l.Wout; WoutNew = [];
%             for i = 1:l.outDim
%                 %Tikhonov Regularization
%                 C = [H;sqrt(l.reg)*eye(l.hidDim)]; 
%                 d = [targets(i,:)';zeros(l.hidDim,1)];
%                 %Constrained Regression   
%                 x = lsqlin(C,d,l.Aieq{i},l.bieq{i},l.Aeq{i},l.beq{i},lb,ub,WoutTemp(:,i),options);
%                 %Assignment
%                 WoutNew(i,:) = x';
%             end
%             l.Wout = WoutNew';
%         end
