
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%          Copyright (c) 2013 K. Neumann                              %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% This example uses the NIVF approach together with a neural Lyapunov candidate

% Please cite following source if you use the neural Lyapunov Function:
% @inproceedings{Neumann2013,
%     author = {K. Neumann and A. Lemme and J. J. Steil},
%     title = {{Neural Learning of Stable Dynamical Systems based on Data-Driven Lyapunov Candidates}},
%     booktitle = {Proc. of the IEEE Int. Conf. on Intelligent Robots and Systems (IROS)},
%     year = {2013},
%     pages={1216--1222}
% }

% Please cite following source if you use neural approaches for learning dynamical systems:
% @inproceedings{Lemme2013,
%     author = {A.~Lemme and K.~Neumann and R.~F.~Reinhart and J.~J.~Steil},
%     title = {{Neurally Imprinted Stable Vector Fields}},
%     booktitle = {Proc. of the Europ. Symp. on Artificial Neural Networks (ESANN)},
%     pages = {327--332},
%     year = {2013}
% }

% Please cite following source if you use neural approaches for learning dynamical systems:
% @article{Lemme2013b,
%     title = {{Neural Learning of Vector Fields for Encoding Stable Dynamical Systems}},
%     journal = {Neurocomputing},
%     note = {In Press},
%     volume = {141}, 
%     pages = {3--14},
%     author = {A.~Lemme and K.~Neumann and R.~F.~Reinhart and J.~J.~Steil}
% }

% Please cite following source if you use the LASA datas set:
% @misc{LASAdataset,
%     author={S. M. Khansari-Zadeh},
%     howpublished={\url{http://www.amarsi-project.eu/open-source}},
%     year={2012},
% }


%% load data
load('../data/DemonstrationsAndLyapunovFunction.mat');

%% data set
X = [PathDemos{1},PathDemos{2},PathDemos{3}];
l1 = diff(PathDemos{1}')'; l2 = diff(PathDemos{2}')'; l3 = diff(PathDemos{3}')';
Y = [diff(PathDemos{1}')',l1(:,end),diff(PathDemos{2}')',l2(:,end),diff(PathDemos{3}')',l3(:,end)];
X = X'; Y = Y';

%% define the sampling region
specs_SamplingRegion = {'class','SR'; 'WS_max',[2,2]; 'WS_min',[-2,-2]};
SR = SamplingRegion(specs_SamplingRegion);

%% setup objective function
specs_ObjFct_X = {'class','CIEQC'; 'partial',{{[1]}}; 'factors',{[-1]}; 'inpDim', 1};
specs_ObjFct_Y = {'class','CIEQC'; 'partial',{{[2]}}; 'factors',{[-1]}; 'inpDim', 1};
objfct = {ObjFct(specs_ObjFct_X),ObjFct(specs_ObjFct_Y)};
        
%% setup continuous inequality constraints
fct_x = @(x) (x(:,1)); fct_y = @(x) (x(:,2));
specs_CIEQC_GRD = {'class','CIEQC'; 'SamplingRegion',SR; 'partial',{{[1]},{[2]}}; 'factors',{fct_x,fct_y}; 'maxValue',Inf; 'minValue',0};
specs_CIEQC_POS = {'class','CIEQC'; 'SamplingRegion',SR; 'partial',{{[]}};        'factors',{[1]};         'maxValue',Inf; 'minValue',0};
CIEQCs = {CIEQC(specs_CIEQC_POS),CIEQC(specs_CIEQC_GRD)};

%% setup discrete equality constraints
specs_DEQC_OT = {'class','DIEQC'; 'partial',{{[]}};  'factors',{[1]}; 'Value',0; 'u',[0,0]};
specs_DEQC_DX = {'class','DIEQC'; 'partial',{{[1]}}; 'factors',{[1]}; 'Value',0; 'u',[0,0]};
specs_DEQC_DY = {'class','DIEQC'; 'partial',{{[2]}}; 'factors',{[1]}; 'Value',0; 'u',[0,0]};
DEQCs = {DEQC(specs_DEQC_OT),DEQC(specs_DEQC_DX),DEQC(specs_DEQC_DY)};

%% network setup
specs_CELM = {'class','CELM'; 'hidDim',100; 'reg',1e-6; 'CIEQCs',CIEQCs; 'DEQCs',DEQCs; 'ObjFct',objfct};
CELM_Lya = CELM(2,1,specs_CELM);

%% learning
CELM_Lya.init(X,Y);
CELM_Lya.train(X,Y);

%% plotting VecField
N_grid = 5000;
xmin = SR.WS_min+0.3; xmax = SR.WS_max-0.3;
[meshX,meshY] = meshgrid(linspace(xmin(1),xmax(1),round(sqrt(N_grid))),linspace(xmin(2),xmax(2),round(sqrt(N_grid))));
InputSeq = makeSeqFromGrid(meshX,meshY);
OutputSeq = CELM_Lya.apply(InputSeq')';
meshZ = makeGridFromSeq(OutputSeq,round(sqrt(N_grid)),round(sqrt(N_grid)));

figure;
    contour(meshX,meshY,meshZ,100);
    hold on;
    plot(PathDemos{1}(1,:),PathDemos{1}(2,:),'.-k');    
    plot(PathDemos{2}(1,:),PathDemos{2}(2,:),'.-k');
    plot(PathDemos{3}(1,:),PathDemos{3}(2,:),'.-k');
    hold off; grid off; box on;
    daspect([1 1 1]);
    axis([xmin(1) xmax(1) xmin(2) xmax(2)]); 
%     xticklabels({}); yticklabels({});
    pause(.1);
    
%% setup continuous inequality constraints
GradLyaFct = @(x) ([CELM_Lya.fctOutput(x,{[1]},[1]),CELM_Lya.fctOutput(x,{[2]},[1])]);
specs_CIEQC = {'class', 'CIEQC'; 'SamplingRegion',SR; 'partial',{{[],[]}}; 'factors',{GradLyaFct}; 'maxValue',0; 'minValue',-Inf};
CIEQCs = {CIEQC(specs_CIEQC)};

%% setup discrete equality constraints
specs_DEQC_X = {'class','DIEQC'; 'partial',{{[],[]}}; 'factors',{[1,0]}; 'Value',0; 'u',[0,0]; 'inpDim', 2};
specs_DEQC_Y = {'class','DIEQC'; 'partial',{{[],[]}}; 'factors',{[0,1]}; 'Value',0; 'u',[0,0]; 'inpDim', 2};
DEQCs = {DEQC(specs_DEQC_X),DEQC(specs_DEQC_Y)};

%% network setup with stability constraints
specs_CELM = {'class','CELM'; 'hidDim',100; 'reg',1e-7; 'CIEQCs',CIEQCs; 'DEQCs',DEQCs}; 
CELM_Dyn = CELM(2,2,specs_CELM);

% learning
CELM_Dyn.init(X,Y);
CELM_Dyn.train(X,Y);

%% network setup without stability constraints
specs_ELM = {'class','CELM'; 'hidDim',100; 'reg',1e-7; 'CIEQCs',{}; 'DEQCs',{}}; 
ELM_Dyn = CELM(2,2,specs_ELM);

% learning
ELM_Dyn.init(X,Y);
ELM_Dyn.train(X,Y);

%% plotting VecField with constraints
N_grid = 5000;
[meshX,meshY] = meshgrid(linspace(xmin(1),xmax(1),round(sqrt(N_grid))),linspace(xmin(2),xmax(2),round(sqrt(N_grid))));
InputSeq = makeSeqFromGrid(meshX,meshY);
OutputSeq = CELM_Dyn.apply(InputSeq');
meshZX = makeGridFromSeq(OutputSeq(:,1)',round(sqrt(N_grid)),round(sqrt(N_grid)));
meshZY = makeGridFromSeq(OutputSeq(:,2)',round(sqrt(N_grid)),round(sqrt(N_grid)));

figure;
    plot(PathDemos{1}(1,:),PathDemos{1}(2,:),'.-k');    
    hold on;
    plot(PathDemos{2}(1,:),PathDemos{2}(2,:),'.-k');
    plot(PathDemos{3}(1,:),PathDemos{3}(2,:),'.-k');
    streamslice(meshX,meshY,meshZX,meshZY);
    scatter(CELM_Dyn.CIEQCs{1,1}.u(:,1), CELM_Dyn.CIEQCs{1,1}.u(:,2))
    hold off; grid off; box on;
    daspect([1 1 1]);
    axis([xmin(1) xmax(1) xmin(2) xmax(2)]); 
    xticklabels({}); yticklabels({});
    title('Learning with Stability Constraints')
    pause(.1);
    
%% plotting VecField without constraints
N_grid = 5000;
[meshX,meshY] = meshgrid(linspace(xmin(1),xmax(1),round(sqrt(N_grid))),linspace(xmin(2),xmax(2),round(sqrt(N_grid))));
InputSeq = makeSeqFromGrid(meshX,meshY);
OutputSeq = ELM_Dyn.apply(InputSeq');
meshZX = makeGridFromSeq(OutputSeq(:,1)',round(sqrt(N_grid)),round(sqrt(N_grid)));
meshZY = makeGridFromSeq(OutputSeq(:,2)',round(sqrt(N_grid)),round(sqrt(N_grid)));

figure;
    plot(PathDemos{1}(1,:),PathDemos{1}(2,:),'.-k');    
    hold on;
    plot(PathDemos{2}(1,:),PathDemos{2}(2,:),'.-k');
    plot(PathDemos{3}(1,:),PathDemos{3}(2,:),'.-k');
    streamslice(meshX,meshY,meshZX,meshZY);
    hold off; grid off; box on;
    daspect([1 1 1]);
    axis([xmin(1) xmax(1) xmin(2) xmax(2)]); 
    xticklabels({}); yticklabels({});
    title('Learning without Stability Constraints')
    pause(.1);
    