
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%          Copyright (c) 2013 K. Neumann                              %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% This example uses the Tau-SEDS approach in combination with the WSAQF
% Lyapunov candidate.

% Please cite following source if you use the Tau-SEDS approach:
% @article{Neumann2015,
%     title = {{Learning Robot Motions with Stable Dynamical Systems under Diffeomorphic Transformations}},
%     journal = {Journal of Robotics and Autonomous Systems},
%     volume = {70},
%     number = {C},
%     pages = {1--15},
%     year = {2015},
%     author = {K. Neumann and and J. J. Steil}
% }

% Please cite following source if you use the neural Lyapunov Function for the Tau-SEDS approach:
% @inproceedings{Neumann2013,
%     author = {K. Neumann and A. Lemme and J. J. Steil},
%     title = {{Neural Learning of Stable Dynamical Systems based on Data-Driven Lyapunov Candidates}},
%     booktitle = {Proc. of the IEEE Int. Conf. on Intelligent Robots and Systems (IROS)},
%     year = {2013},
%     pages={1216--1222}
% }

% Please cite following source if you use neural approaches for learning dynamical systems:
% @inproceedings{Lemme2013,
%     author = {A.~Lemme and K.~Neumann and R.~F.~Reinhart and J.~J.~Steil},
%     title = {{Neurally Imprinted Stable Vector Fields}},
%     booktitle = {Proc. of the Europ. Symp. on Artificial Neural Networks (ESANN)},
%     pages = {327--332},
%     year = {2013}
% }

% Please cite following source if you use neural approaches for learning dynamical systems:
% @article{Lemme2013b,
%     title = {{Neural Learning of Vector Fields for Encoding Stable Dynamical Systems}},
%     journal = {Neurocomputing},
%     note = {In Press},
%     volume = {141}, 
%     pages = {3--14},
%     author = {A.~Lemme and K.~Neumann and R.~F.~Reinhart and J.~J.~Steil}
% }

% Please cite following source if you use the SEDS Software:
% @misc{SEDSSoftware,
%     author={S. M. Khansari-Zadeh},
%     howpublished={\url{http://lasa.epfl.ch/people/member.php?SCIPER=183746/}},
%     year={2013},
% }

% Please cite following source if you use the LASA datas set:
% @misc{LASAdataset,
%     author={S. M. Khansari-Zadeh},
%     howpublished={\url{http://www.amarsi-project.eu/open-source}},
%     year={2012},
% }

% Please cite following source if you use the SEDS approach:
% @article{KhansariZadeh2011,
%     title={{Learning Stable Non-Linear Dynamical Systems with Gaussian Mixture Models}},
%     author={Khansari-Zadeh, S. M. and Billard, A.},
%     journal={Trans. on Robotics},
%     volume={27},
%     number={5},
%     pages={943--957},
%     year={2011}
% }

% Please cite following source if you use the WSAQF and SEDS aproach:
% @article{KhansariZadeh2014,
%     title={{Learning control lyapunov function to ensure stability of dynamical system-based robot reaching motions}},
%     author={Khansari-Zadeh, S. M. and Billard, A.},
%     journal={Trans. on Robotics},
%     volume={62},
%     number={6},
%     pages={752-765},
%     year={2014}
% }

%% load data
load('../data/DemonstrationsAndLyapunovFunction.mat');

%% data set
N_demos = 3;
X = [PathDemos{1},PathDemos{2},PathDemos{3}];
l1 = diff(PathDemos{1}')'; l2 = diff(PathDemos{2}')'; l3 = diff(PathDemos{3}')';
Y = [diff(PathDemos{1}')',l1(:,end),diff(PathDemos{2}')',l2(:,end),diff(PathDemos{3}')',l3(:,end)];
X = X'; Y = Y';

%% WSAQF
specs_WSAQF = {'class','WSAQF'; 'num_L',2; 'MaxFunEvals',1000; 'reg',1; 'w',1};
wsaqf = WSAQF(2,1,specs_WSAQF);
wsaqf.init();
wsaqf.train(X,Y);

%% transformation setup
specs_TSEDS = {'class','TSEDS'; 'LyaFct',@(x) ( wsaqf.apply(x) ); 'GradLyaFct',@(x) (wsaqf.gradient(x))}; 
tseds = TSEDS(2,specs_TSEDS);
tseds.init(X);                   % init
[Xn,Yn] = tseds.CoordTrafo(X,Y); % data trafo

%% learning SEDS 
options.tol_mat_bias = 10^-6;  % A very small positive scalar to avoid instabilities in Gaussian kernel [default: 10^-15]                            
options.display = 1;           % An option to control whether the algorithm displays the output of each iterations [default: true]
options.tol_stopping=10^-10;   % A small positive scalar defining the stoppping tolerance for the optimization solver [default: 10^-10]
options.max_iter = 500;        % Maximum number of iteration for the solver [default: i_max=1000]
options.objective = 'mse';     % 'likelihood': use likelihood as criterion to optimize parameters of GMM
options.normalization = 1;     % data normalization
K = 5;                         % number of Gaussian functions
d = tseds.dim;                 % dimensionality of the task
Data = [Xn';Yn'];              % transformed data to learn for SEDS

[Priors_0, Mu_0, Sigma_0] = initialize_SEDS(Data,K);                 %finding an initial guess for GMM's parameter
[Priors,Mu,Sigma] = SEDS_Solver(Priors_0,Mu_0,Sigma_0,Data,options); %running SEDS optimization solver
tseds.VecField = @(x) GMR(Priors,Mu,Sigma,x,1:d,d+1:2*d);

%% plotting WSAQF
N_grid = 5000;
xmin(1) = min(X(:,1)); xmax(1) = max(X(:,1)); 
xmin(2) = min(X(:,2)); xmax(2) = max(X(:,2));
[meshX,meshY] = meshgrid(linspace(xmin(1),xmax(1),round(sqrt(N_grid))),linspace(xmin(2),xmax(2),round(sqrt(N_grid))));
InputSeq = makeSeqFromGrid(meshX,meshY);
out = log(wsaqf.apply(InputSeq'));

figure;    
    contour(meshX,meshY,makeGridFromSeq(out,round(sqrt(N_grid)),round(sqrt(N_grid))),100);
    hold on
    for i = 1:N_demos
        plot(PathDemos{i}(1,:),PathDemos{i}(2,:),'-k.','Linewidth',1.5,'MarkerSize',12);
    end
    hold off; grid off; box on;
    daspect([1 1 1]);
    axis([xmin(1) xmax(1) xmin(2) xmax(2)]);    
    title('WSAQF Lyapunov Function in Original Space');
    
%% plotting L=x^2
N_grid = 5000;
xmin(1) = min(Xn(:,1)); xmax(1) = max(Xn(:,1)); 
xmin(2) = min(Xn(:,2)); xmax(2) = max(Xn(:,2));
[meshX,meshY] = meshgrid(linspace(xmin(1),xmax(1),round(sqrt(N_grid))),linspace(xmin(2),xmax(2),round(sqrt(N_grid))));
InputSeq = makeSeqFromGrid(meshX,meshY);
out = sum(InputSeq.^2);

figure;    
    contour(meshX,meshY,makeGridFromSeq(out,round(sqrt(N_grid)),round(sqrt(N_grid))),100);
    hold on
    for i = 1:N_demos
        plot(Xn(:,1),Xn(:,2),'.k','Linewidth',1.5,'MarkerSize',12);
    end
    hold off; grid off; box on;
    daspect([1 1 1]);
    axis([xmin(1) xmax(1) xmin(2) xmax(2)]);  
    title('WSAQF Lyapunov Function in Transformed Space');
   
%% plotting vector field
for i = 1:size(InputSeq,2)
    outs_VF(:,i) = tseds.apply_notrafo(InputSeq(:,i)');
end  
outsX_VF = makeGridFromSeq(outs_VF(1,:),round(sqrt(N_grid)),round(sqrt(N_grid)));
outsY_VF = makeGridFromSeq(outs_VF(2,:),round(sqrt(N_grid)),round(sqrt(N_grid)));

figure;    
    streamslice(meshX,meshY,outsX_VF,outsY_VF);
    hold on
    plot(Xn(:,1),Xn(:,2),'.k','Linewidth',1.5,'MarkerSize',12);
    hold off; grid off; box on;
    daspect([1 1 1]);
    axis([xmin(1) xmax(1) xmin(2) xmax(2)]); 
    title('SEDS Vector Field in Transformed Space');
    
%% plotting vector field
xmin(1) = min(X(:,1)); xmax(1) = max(X(:,1)); 
xmin(2) = min(X(:,2)); xmax(2) = max(X(:,2));
[meshX,meshY] = meshgrid(linspace(xmin(1),xmax(1),round(sqrt(N_grid))),linspace(xmin(2),xmax(2),round(sqrt(N_grid))));
InputSeq = makeSeqFromGrid(meshX,meshY);
for i = 1:size(InputSeq,2)
    outs_VF(:,i) = tseds.apply(InputSeq(:,i)');
end  
outsX_VF = makeGridFromSeq(outs_VF(1,:),round(sqrt(N_grid)),round(sqrt(N_grid)));
outsY_VF = makeGridFromSeq(outs_VF(2,:),round(sqrt(N_grid)),round(sqrt(N_grid)));

figure;    
    streamslice(meshX,meshY,outsX_VF,outsY_VF);
    hold on
    plot(X(:,1),X(:,2),'.k','Linewidth',1.5,'MarkerSize',12);
    hold off; grid off; box on;
    daspect([1 1 1]);
    axis([xmin(1) xmax(1) xmin(2) xmax(2)]);   
    title('SEDS Vector Field in Original Space');

